from __future__ import annotations

import re
import struct
import warnings
from enum import IntEnum
from typing import BinaryIO

import numpy as np

from .commons import ByteSize, IdefixFieldProperties, IdefixMetadata

KNOWN_GEOMETRIES: dict[int, str] = {
    1: "cartesian",
    2: "cylindrial",
    3: "polar",
    4: "spherical",
}


class CharCount(IntEnum):
    # hardcoded in idefix
    HEADER = 128
    NAME = 16


# emulating C++
# enum DataType {DoubleType, SingleType, IntegerType};
DTYPES = {0: "d", 1: "f", 2: "i"}
DTYPES_2_NUMPY = {"d": "=f8", "f": "=f4", "i": "=i4"}
DTYPES_2_SIZE = {"i": ByteSize.INT, "f": ByteSize.FLOAT, "d": ByteSize.DOUBLE}


def read_null_terminated_string(fh: BinaryIO, maxsize: int = CharCount.NAME):
    """Read maxsize * ByteSize.CHAR bytes, but only parse non-null characters."""
    b = fh.read(maxsize * ByteSize.CHAR)
    s = b.decode("utf-8", errors="backslashreplace")
    s = s.split("\x00", maxsplit=1)[0]
    return s


def read_next_field_properties(fh: BinaryIO):
    """Emulate Idefix's OutputDump::ReadNextFieldProperty"""
    field_name = read_null_terminated_string(fh)

    fmt = "=i"
    dtype = DTYPES[struct.unpack(fmt, fh.read(struct.calcsize(fmt)))[0]]
    ndim = struct.unpack(fmt, fh.read(struct.calcsize(fmt)))[0]
    if ndim > 3:
        raise ValueError(ndim)
    fmt = f"={ndim}i"
    dim = np.array(struct.unpack(fmt, fh.read(struct.calcsize(fmt))))
    return field_name, dtype, ndim, dim


def read_chunk(
    fh: BinaryIO,
    ndim: int,
    dim: list[int],
    dtype: str,
    *,
    is_scalar: bool = False,
    skip_data: bool = False,
) -> np.ndarray | None:
    # NOTE: ret type is only dependent on skip_data...
    # this could be better expressed in the type annotationation but it would make
    # more sense to just refactor this function to avoid the boolean trap, so I'll keep wonky
    # type hints for now
    assert ndim == len(dim)
    count = np.product(dim)
    size = count * DTYPES_2_SIZE[dtype]
    if skip_data:
        fh.seek(size, 1)
        return None

    # note: this reversal may not be desirable in general
    if is_scalar:
        fmt = f"={count}{dtype}"
        retv = struct.unpack(fmt, fh.read(size))[0]
        return retv
    else:
        data = np.fromfile(fh, DTYPES_2_NUMPY[dtype], count=count)
        data.shape = dim[::-1]
        return data.T


def read_serial(
    fh: BinaryIO, ndim: int, dim: list[int], dtype: str, *, is_scalar: bool = False
) -> np.ndarray | None:
    """Emulate Idefix's OutputDump::ReadSerial"""
    assert ndim == 1  # corresponds to an error raised in IDEFIX
    return read_chunk(fh, ndim=ndim, dim=dim, dtype=dtype, is_scalar=is_scalar)


def read_distributed(
    fh: BinaryIO, dim: list[int], *, skip_data: bool = False
) -> np.ndarray | None:
    """Emulate Idefix's OutputDump::ReadDistributed"""
    # note: OutputDump::ReadDistributed only read doubles
    # because chucks written in integers are small enough
    # that parallelization is counter productive.
    # This a design choice on idefix's size.
    return read_chunk(fh, ndim=len(dim), dim=dim, dtype="d", skip_data=skip_data)


# The following functions are originally designed for yt


def read_header(filename: str) -> str:
    with open(filename, "rb") as fh:
        header = read_null_terminated_string(fh, maxsize=CharCount.HEADER)
    return header


def get_field_offset_index(fh: BinaryIO) -> dict[str, int]:
    """
    Go over a dumpfile, parse bytes offsets associated with each field.
    Returns
    -------
    field_index: mapping (field name -> offset)
    """
    field_index = {}

    # skip header
    fh.seek(CharCount.HEADER * ByteSize.CHAR)
    # skip grid properties
    for _ in range(9):
        _field_name, dtype, ndim, dim = read_next_field_properties(fh)
        read_serial(fh, ndim, dim, dtype)

    while True:
        offset = fh.tell()
        field_name, dtype, ndim, dim = read_next_field_properties(fh)
        if not re.match("^V[cs]-", field_name):
            break
        field_index[field_name] = offset
        read_distributed(fh, dim, skip_data=True)

    return field_index


def read_single_field(fh: BinaryIO, field_offset: int) -> np.ndarray:
    """
    Returns
    -------
    data: 3D np.ndarray with dtype float64
    """
    fh.seek(field_offset)
    field_name, dtype, ndim, dim = read_next_field_properties(fh)
    data = read_distributed(fh, dim)
    return data


def read_idefix_dmpfile(
    filename: str, skip_data: bool = False
) -> tuple[IdefixFieldProperties, IdefixMetadata]:
    with open(filename, "rb") as fh:
        return read_idefix_dump_from_buffer(fh, skip_data)


def read_idefix_dump_from_buffer(
    fh: BinaryIO, skip_data: bool = False
) -> tuple[IdefixFieldProperties, IdefixMetadata]:

    # skip header
    fh.seek(CharCount.HEADER * ByteSize.CHAR)

    fprops = {}
    fdata = {}
    for _ in range(9):
        # read grid properties
        # (cell centers, left and right edges in 3D -> 9 arrays)
        field_name, dtype, ndim, dim = read_next_field_properties(fh)
        data = read_serial(fh, ndim, dim, dtype)
        fprops[field_name] = dtype, ndim, dim
        fdata[field_name] = data

    field_name, dtype, ndim, dim = read_next_field_properties(fh)
    while field_name != "eof":
        # note that this could likely be implemented using a call to
        # `iter` with a sentinel value, to the condition that read_next_field_properties
        # would be splitted into 2 parts (I don't the sentinel pattern works with tuples)
        fprops[field_name] = dtype, ndim, dim
        if field_name.startswith("Vc-") or field_name.startswith("Vs-"):
            data = read_distributed(fh, dim, skip_data=skip_data)
        else:
            is_scalar = ndim == 1 and dim[0] == 1
            is_scalar &= field_name not in ("x1", "x2", "x3")
            data = read_serial(fh, ndim, dim, dtype, is_scalar=is_scalar)
        fdata[field_name] = data
        field_name, dtype, ndim, dim = read_next_field_properties(fh)

    if "geometry" not in fdata:
        warnings.warn("Could not determine geometry, assuming cartesian.")
        fdata["geometry"] = "cartesian"
    elif isinstance(fdata["geometry"], int):
        fdata["geometry"] = KNOWN_GEOMETRIES[fdata["geometry"]]
    else:
        raise RuntimeError(
            f"Got a geometry flag with unexpected type {type(fdata['geometry'])}; "
            f"got {fdata['geometry']}"
        )

    return fprops, fdata



from __future__ import annotations

import struct
import warnings
from typing import Any, BinaryIO

import numpy as np

from .commons import ByteSize, Coordinates, Shape

KNOWN_GEOMETRIES = {0: "cartesian", 1: "polar", 2: "spherical"}


def read_header(filename: str) -> str:
    with open(filename, "rb") as fh:
        return "".join(fh.readline(256).decode() for _ in range(2))


def read_single_field(
    fh: BinaryIO,
    offset: int | None = None,
    *,
    shape: tuple[int, int, int],
    skip_data: bool = False,
) -> np.ndarray | None:
    count = np.prod(shape)
    if offset is not None and fh.tell() != offset:
        fh.seek(offset)
    if skip_data:
        fh.seek(count * ByteSize.FLOAT, 1)
        data = None
    else:
        data = np.fromfile(fh, ">f", count=count)
        data.shape = shape[::-1]
        data = data.T
    return data


def read_shape(s: str) -> Shape:
    # read a specific line containing nx, ny, nz

    assert s.startswith("DIMENSIONS")
    raw: list[str] = s.split()[1:]
    if len(raw) != 3:
        raise RuntimeError
    return Shape(*(int(_) for _ in raw))


def parse_shape(s: str, md: dict[str, Any]) -> None:
    md["shape"] = read_shape(s)


# this may not be kept in the following form
def read_metadata(fh: BinaryIO) -> dict[str, Any]:

    fh.seek(0)
    # skip over the first 4 lines which normally contains
    # VTK DataFile Version x.x
    # <Comments>
    # BINARY
    # DATASET RECTILINEAR_GRID or STRUCTURED_GRID
    for _ in range(4):
        next(fh)

    metadata = {}
    line = next(fh).decode()  # DIMENSIONS NX NY NZ or FIELD
    if line.startswith("FIELD"):
        # Idefix >= 0.8
        nfield = int(line.split()[2])
        for _ in range(nfield):
            d = next(fh).decode()
            if d.startswith("GEOMETRY"):
                geom_flag: int = struct.unpack(">i", fh.read(4))[0]
                metadata["geometry"] = KNOWN_GEOMETRIES.get(geom_flag, "unknown")
            elif d.startswith("TIME"):
                metadata["time"] = struct.unpack(">f", fh.read(4))[0]
            else:
                warnings.warn(f"Found unknown field {d!r}")
            next(fh)  # skip extra linefeed (empty line)
        parse_shape(next(fh).decode(), metadata)

    elif line.startswith("DIMENSIONS"):
        # Idefix < 0.8
        parse_shape(line, metadata)
        warnings.warn(
            "yt_idefix has limited support for vtk files with idefix < 0.8 . "
            "Assuming a cartesian geometry."
        )
        metadata["geometry"] = "cartesian"

    else:
        raise RuntimeError(f"Failed to parse {line!r}")

    return metadata


def read_grid_coordinates(
    fh: BinaryIO, md: dict[str, Any] | None = None
) -> Coordinates:
    if md is None:
        fh.seek(0)
        md = read_metadata(fh)

    coords: list[np.ndarray] = []
    # now assuming that fh is positioned at the end of metadata
    if md["geometry"] == "cartesian":
        for key in ("n1", "n2", "n3"):
            next(fh)
            coords.append(np.fromfile(fh, dtype=">f", count=getattr(md["shape"], key)))
            next(fh)
        line = next(fh).decode()
        point_type, npoints = (t(_) for t, _ in zip((str, int), line.split()))
        next(fh)
        if point_type == "CELL_DATA":
            # raw data is cell face coords
            # we'd need to interpolate cell centers
            warnings.warn(
                "point_type = CELL_DATA is not fully supported yet. "
                "This will be treated as POINT_DATA instead"
            )
            md["array_shape"] = md["shape"].to_cell_centered()
        else:
            md["array_shape"] = md["shape"]

    else:
        raise NotImplementedError("Non cartesian coordinates are not implemented yet")

    return Coordinates(*coords)


def read_field_offset_index(fh: BinaryIO, shape: Shape) -> dict[str, int]:
    # assuming fh is correctly positioned (read_grid_coordinates must be called first)
    retv: dict[str, int] = {}

    while True:
        line = fh.readline()
        if len(line) < 2:
            break
        s = line.decode()
        datatype, varname, dtype = s.split()

        if datatype == "SCALARS":
            next(fh)
            retv[varname] = fh.tell()
            read_single_field(fh, shape=shape, skip_data=True)
        elif datatype == "VECTORS":
            for axis in "XYZ":
                vname = f"{varname}_{axis}"
                retv[vname] = fh.tell()
                read_single_field(fh, shape=shape, skip_data=True)
        else:
            raise RuntimeError(f"Unknown datatype {datatype!r}")
        fh.readline()
    return retv
